import numpy as np

def CA(reports, agent_first = True):
	reports = np.array(reports)
	print(reports.shape)
	if agent_first:
		reports = reports.transpose()
	m, n = reports.shape # m agents n tasks
	options = np.unique(reports)
	option_index = {}
	for key, item in enumerate(options):
		option_index[item] = key

	#separate tasks
	groupA = np.random.choice(m, m // 2, replace = False)
	groupB = np.setdiff1d(np.arange(0, m), groupA)
	def learning(reports):
		k = options.shape[0]
		joint = np.zeros([k, k], dtype = float)
		alpha = 1.0 / reports.shape[0]
		for task in reports:
			sample = np.random.choice(task, 2, replace = False)
			i = option_index[sample[0]]
			j = option_index[sample[1]]
			joint[i][j] += alpha
		marginal = joint.sum(axis = 1)
		return joint - np.matmul(marginal.reshape(k, 1), marginal.reshape(1, k))

	DeltaA = learning(reports[groupA, :]) > 0
	DeltaB = learning(reports[groupB, :]) > 0
	print(learning(reports[groupA, :]), learning(reports[groupB, :]))

	score = np.zeros([n], dtype = int)
	def scorer(reports, scoreMatrix):
		for task in reports:
			for agent1 in range(n):
				for agent2 in range(agent1):
					report1 = option_index[task[agent1]]
					report2 = option_index[task[agent2]]
					if scoreMatrix[report1][report2]:
						score[agent1] += 1
						score[agent2] += 1
		return 

	scorer(reports[groupA, :], DeltaB)
	scorer(reports[groupB, :], DeltaA)
	return score

if __name__ == '__main__':
	reports = np.array([np.random.choice(['subway', 'burgerK', 'McDonald', 'KFC', 'PizzaHot'], 100, p=[0.1, 0.2, 0.2, 0, 0.5]) for x in range(8)])
	score = CA(reports)
	print(score)


import numpy as np
import math

def comb(n, m):
    return math.factorial(n) / (math.factorial(m) * math.factorial(n - m))

def dmi2(A1, B1, A2, B2, C):
    def check(x):
        return 0 <= x and x < C and type(x) == np.int64

    def GetM(A, B):
        assert(len(A) == len(B))
        M = np.zeros((C, C))
        for (x, y) in zip(A, B):
            if check(x) and check(y):
                M[x][y] += 1
            else:
                raise ValueError('The values of answers must be integers in [0, C)')
        return M

    M1 = GetM(A1, B1)
    M2 = GetM(A2, B2)

    return np.linalg.det(M1) * np.linalg.det(M2)

class functions:
    def __init__(self, a, b, c):
        self.f = a
        self.f_prime = b
        self.f_star = c

def TVD(t):
    return abs(t - 1)
def TVD_prime(x):
    x = math.log(x)
    if x > 0: return 1
    elif x == 0: return 0
    else: return -1
def TVD_star(x):
    return TVD_prime(x)
TVDs = functions(TVD, TVD_prime, TVD_star)

def KLD(t):
    return t * math.log(t)
def KLD_prime(t):
    return 1 + math.log(t)
def KLD_star(t):
    return t
KLDs = functions(KLD, KLD_prime, KLD_star)


def R(f, a, b, p):
    return f( a*b/p + (1-a)*(1-b)/(1-p) )

def MIG(answers, function, prior):
    agent_n, task_n = answers.shape

    # N > 1;
    if agent_n != 2:
        raise ValueError('Invalid number of agents.')

    # f
    fs = TVDs
    if function == 'KLD':
        fs = KLDs

    # Calculate reward
    reward = 0
    penalty = 0
    for i in range(task_n):
        for j in range(task_n):
            if i == j:
                reward += R(fs.f_prime, answers[0][i], answers[1][i], prior)
            else:
                penalty += R(fs.f_star, answers[0][i], answers[1][j], prior)
    payment = reward / task_n - penalty / (task_n * (task_n - 1))

    return [payment, payment]


if __name__ == '__main__':
    arr = np.array([[0.2, 0.3, 0.2], [0.3, 0.5, 0.3]])
    print(MIG(arr, 'TVD', 0.3))


import numpy as np

def BTS(x, y, alpha = 1.0):
	x, y = np.array(x), np.array(y)
	n, m = y.shape

	_xbar = np.zeros([m], dtype = float)
	_ybar = np.zeros([m], dtype = float)
	for opinion in x:
		_xbar[opinion] += 1.0
	for player in y:
		_ybar += np.log(player)
	def xbar(k, r, s):
		#return _xbar[k] / n
		return (_xbar[k] + 1 - float(x[r] == k) - float(x[s] == k)) / (n + m - 2)
	def ybar(k, r, s):
		#return _ybar[k] / n
		return (_ybar[k] - np.log(y[r][k]) - np.log(y[s][k])) / (n - 2)

	score = np.zeros([n], dtype = float)
	for r in range(n):
		info, pred = 0.0, 0.0
		for s in range(n):
			if r == s: continue
			info += np.log(xbar(x[r], r, s)) - ybar(x[r], r, s)
			for k in range(m):
				pred += xbar(k, r, s) * np.log(y[r][k] / xbar(k, r, s))
		score[r] = info + pred * alpha
	return score

score = BTS([0 for i in range(100 - 1)] + [1], [[0.99, 0.01] for i in range(100)])
print(score.sum())

score = BTS([3, 2, 1, 1, 0],
[[0.1, 0.1, 0.3, 0.5],
[0.1, 0.2, 0.5, 0.2],
[0.3, 0.4, 0.2, 0.1],
[0.3, 0.4, 0.1, 0.2],
[0.1, 0.3, 0.2, 0.4]])
print(score, score.sum())


import numpy as np
import math

def comb(n, m):
    return math.factorial(n) / (math.factorial(m) * math.factorial(n - m))

def dmi2(A1, B1, A2, B2, C):
    def check(x):
        return 0 <= x and x < C and type(x) == np.int64

    def GetM(A, B):
        assert(len(A) == len(B))
        M = np.zeros((C, C))
        for (x, y) in zip(A, B):
            if check(x) and check(y):
                M[x][y] += 1
            else:
                raise ValueError('The values of answers must be integers in [0, C)')
        return M

    M1 = GetM(A1, B1)
    M2 = GetM(A2, B2)

    return np.linalg.det(M1) * np.linalg.det(M2)


def DMI(answers, choice_n):
    agent_n, task_n = answers.shape

    # T >= 2C; N > 1;
    if task_n < 2 * choice_n:
        raise ValueError('Insufficient number of tasks.')
    if agent_n <= 1:
        raise ValueError('Too few agents.')

    # T tasks are arbitrarily divided into two disjoint parts T_1 , T_2
    answers = np.transpose(answers)
    np.random.shuffle(answers)
    half = task_n // 2
    T1 = np.transpose(answers[ : half])
    T2 = np.transpose(answers[half : ])

    # Calculate the payment
    payments = []
    norm_factor = (agent_n - 1) * (math.factorial(choice_n) ** 2)
    norm_factor *= comb(T1.shape[0], choice_n) * comb(T2.shape[0], choice_n)
    for i in range(agent_n):
        p = 0
        for j in range(agent_n):
            if i == j: continue
            p += dmi2(T1[i], T1[j], T2[i], T2[j], choice_n)
        p /= norm_factor
        payments.append(p)

    return payments


if __name__ == '__main__':
    arr = np.array([[1, 1, 0, 1, 1, 0, 1], [1, 1, 0, 0, 1, 0, 1]])
    print(DMI(arr, 2))